/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.common.attached_item;

import static java.util.Objects.*;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import java.util.function.Function;
import java.util.stream.Collectors;

import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Service;

import de.ozgcloud.apilib.common.command.OzgCloudCommand;
import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.apilib.common.command.grpc.CommandMapper;
import de.ozgcloud.bescheid.Bescheid;
import de.ozgcloud.bescheid.Bescheid.Status;
import de.ozgcloud.bescheid.BescheidCallContextAttachingInterceptor;
import de.ozgcloud.bescheid.BescheidManagerConfiguration;
import de.ozgcloud.bescheid.common.freemarker.TemplateHandler;
import de.ozgcloud.bescheid.vorgang.VorgangId;
import de.ozgcloud.command.Command;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.document.Document;
import lombok.RequiredArgsConstructor;

@Service
@RequiredArgsConstructor
public class AttachedItemService {

	public static final String BESCHEID_ITEM_NAME = "Bescheid";
	public static final String PATCH_ATTACHED_ITEM = "PATCH_ATTACHED_ITEM";
	static final String CREATE_ATTACHED_ITEM_ORDER = "CREATE_ATTACHED_ITEM";
	static final String UPDATE_ATTACHED_ITEM_ORDER = "UPDATE_ATTACHED_ITEM";
	static final String DELETE_ATTACHED_ITEM = "DELETE_ATTACHED_ITEM";

	static final String DEFAULT_SUBJECT = "Ihr Bescheid zum Antrag";
	static final String DEFAULT_TEMPLATE_FILE = "bescheid.nachrichtTemplate.txt.ftlh";

	@Qualifier(BescheidManagerConfiguration.COMMAND_SERVICE_NAME)
	private final OzgCloudCommandService commandService;
	private final VorgangAttachedItemRemoteService remoteService;
	private final CommandMapper commandMapper;
	private final TemplateHandler templateHandler;

	public Optional<AttachedItem> findBescheidItem(String vorgangId) {
		return remoteService.findBescheidDraft(vorgangId);
	}

	public AttachedItem getItem(String id) {
		return remoteService.getItem(id);
	}

	public String createBescheidDraft(Command command) {
		validateBescheidData(command.getBodyObject());
		return remoteService.findBescheidDraft(command.getVorgangId())
				.map(bescheid -> overrideAttachedItem(bescheid, command)).orElseGet(() -> createAttachedItem(command));
	}

	void validateBescheidData(Map<String, Object> bodyObject) {
		validateBescheidField(bodyObject, Bescheid.FIELD_BESCHIEDEN_AM);
		validateBescheidField(bodyObject, Bescheid.FIELD_BEWILLIGT);
		validateSendBy(getSendBy(bodyObject));
	}

	void validateBescheidField(Map<String, Object> bescheid, String field) {
		if (isNull(bescheid.get(field))) {
			throw new TechnicalException("Fields '%s' is required for bescheid creation".formatted(field));
		}
	}

	void validateSendBy(String sendBy) {
		if (isUnknownSendByValue(sendBy)) {
			var possibleSendByValues = Arrays.stream(Bescheid.SendBy.values()).map(Bescheid.SendBy::name)
					.collect(Collectors.joining(","));
			var message = String.format("Unexpected value for field '%s': %s. Allowed are: %s", Bescheid.FIELD_SEND_BY, sendBy,
					possibleSendByValues);
			throw new TechnicalException(message);
		}
	}

	private boolean isUnknownSendByValue(String sendByValue) {
		return Arrays.stream(Bescheid.SendBy.values()).noneMatch(sendBy -> sendBy.hasValue(sendByValue));
	}

	String overrideAttachedItem(AttachedItem bescheidItem, Command command) {
		var attachedItem = buildAttachedItemAsMap(command, buildItemMapWithAllBescheidFields(command));
		var finishedOzgCloudCommand = commandService.createAndWaitUntilDone(buildPatchAttachedItemCommand(bescheidItem, attachedItem));
		return finishedOzgCloudCommand.getCreatedResource();
	}

	Map<String, Object> buildItemMapWithAllBescheidFields(Command command) {
		var result = new HashMap<String, Object>();
		result.put(Bescheid.FIELD_BESCHIEDEN_AM, MapUtils.getObject(command.getBodyObject(), Bescheid.FIELD_BESCHIEDEN_AM));
		result.put(Bescheid.FIELD_BEWILLIGT, MapUtils.getObject(command.getBodyObject(), Bescheid.FIELD_BEWILLIGT));
		result.put(Bescheid.FIELD_BESCHEID_DOCUMENT,
				MapUtils.getString(command.getBodyObject(), Bescheid.FIELD_BESCHEID_DOCUMENT));
		result.put(Bescheid.FIELD_SEND_BY, getSendBy(command.getBodyObject()));
		result.put(Bescheid.FIELD_NACHRICHT_TEXT, MapUtils.getString(command.getBodyObject(), Bescheid.FIELD_NACHRICHT_TEXT));
		result.put(Bescheid.FIELD_NACHRICHT_SUBJECT,
				MapUtils.getString(command.getBodyObject(), Bescheid.FIELD_NACHRICHT_SUBJECT));
		result.put(Bescheid.FIELD_ATTACHMENTS, MapUtils.getObject(command.getBodyObject(), Bescheid.FIELD_ATTACHMENTS));
		return result;
	}

	private String getSendBy(Map<String, Object> body) {
		return MapUtils.getString(body, Bescheid.FIELD_SEND_BY);
	}

	public void updateBescheidDraft(Command command, Optional<Document> document) {
		createCommandAfterValidation(command, bescheid -> buildUpdateBescheidDraftCommand(command, document, bescheid));
	}

	private OzgCloudCommand buildUpdateBescheidDraftCommand(Command command, Optional<Document> document, AttachedItem bescheidItem) {
		return buildPatchAttachedItemCommand(bescheidItem, buildAttachedItemAsMap(command, buildBescheidItemAsMap(command, document)));
	}

	Map<String, Object> buildBescheidItemAsMap(Command command, Optional<Document> newDocument) {
		var itemMap = buildItemMapWithExistingBescheidFields(command);
		var savedBescheid = getItem(command.getRelationId());

		getNewNachrichtText(savedBescheid, newDocument)
				.ifPresent(nachrichtText -> itemMap.put(Bescheid.FIELD_NACHRICHT_TEXT, nachrichtText));

		return itemMap;
	}

	Optional<String> getNewNachrichtText(AttachedItem bescheidItem, Optional<Document> newDocument) {
		return newDocument
				.filter(document -> isDifferentDocument(document, bescheidItem))
				.map(Document::getNachrichtText)
				.filter(StringUtils::isNotBlank);
	}

	boolean isDifferentDocument(Document document, AttachedItem bescheidItem) {
		return !StringUtils.equals(document.getId(), getBescheidDocumentId(bescheidItem));
	}

	private String getBescheidDocumentId(AttachedItem item) {
		return MapUtils.getString(item.getItem(), Bescheid.FIELD_BESCHEID_DOCUMENT);
	}

	OzgCloudCommand buildPatchAttachedItemCommand(AttachedItem bescheidItem, Map<String, Object> bodyObject) {
		return OzgCloudCommand.builder()
				.vorgangId(commandMapper.toOzgCloudVorgangId(bescheidItem.getVorgangId()))
				.relationId(commandMapper.mapRelationId(bescheidItem.getId()))
				.relationVersion(bescheidItem.getVersion())
				.order(PATCH_ATTACHED_ITEM)
				.bodyObject(bodyObject)
				.build();
	}

	String createAttachedItem(Command command) {
		return commandService.createAndWaitUntilDone(buildCreateAttachedItemCommand(command)).getCreatedResource();
	}

	OzgCloudCommand buildCreateAttachedItemCommand(Command command) {
		return OzgCloudCommand.builder()
				.vorgangId(commandMapper.toOzgCloudVorgangId(command.getVorgangId()))
				.relationId(commandMapper.mapRelationId(command.getRelationId()))
				.order(CREATE_ATTACHED_ITEM_ORDER)
				.bodyObject(buildAttachedItemAsMap(command, buildCreateBescheidItemMap(command))).build();
	}

	Map<String, Object> buildAttachedItemAsMap(Command command, Map<String, Object> itemMap) {
		var result = new HashMap<String, Object>();
		result.put(AttachedItem.PROPERTY_VORGANG_ID, command.getVorgangId());
		result.put(AttachedItem.PROPERTY_CLIENT, BescheidCallContextAttachingInterceptor.BESCHEID_MANAGER_CLIENT_NAME);
		result.put(AttachedItem.PROPERTY_ITEM_NAME, BESCHEID_ITEM_NAME);
		result.put(AttachedItem.PROPERTY_ITEM, itemMap);
		return result;
	}

	Map<String, Object> buildCreateBescheidItemMap(Command command) {
		var result = buildItemMapWithExistingBescheidFields(command);
		if (StringUtils.isEmpty(getNachrichtSubject(command))) {
			result.put(Bescheid.FIELD_NACHRICHT_SUBJECT, DEFAULT_SUBJECT);
		}
		if (StringUtils.isEmpty(getNachrichtText(command))) {
			result.put(Bescheid.FIELD_NACHRICHT_TEXT, templateHandler.getRawTemplate(DEFAULT_TEMPLATE_FILE));
		}
		return result;
	}

	Map<String, Object> buildItemMapWithExistingBescheidFields(Command command) {
		var result = buildItemMap(Bescheid.Status.DRAFT);
		addValueFromMapIfExists(command.getBodyObject(), Bescheid.FIELD_BESCHIEDEN_AM, result);
		addValueFromMapIfExists(command.getBodyObject(), Bescheid.FIELD_BEWILLIGT, result);
		addValueFromMapIfExists(command.getBodyObject(), Bescheid.FIELD_BESCHEID_DOCUMENT, result);
		addValueFromMapIfExists(command.getBodyObject(), Bescheid.FIELD_SEND_BY, result);
		addValueFromMapIfExists(command.getBodyObject(), Bescheid.FIELD_NACHRICHT_TEXT, result);
		addValueFromMapIfExists(command.getBodyObject(), Bescheid.FIELD_NACHRICHT_SUBJECT, result);
		addValueFromMapIfExists(command.getBodyObject(), Bescheid.FIELD_ATTACHMENTS, result);
		return result;
	}

	private String getNachrichtSubject(Command command) {
		return MapUtils.getString(command.getBodyObject(), Bescheid.FIELD_NACHRICHT_SUBJECT);
	}

	private String getNachrichtText(Command command) {
		return MapUtils.getString(command.getBodyObject(), Bescheid.FIELD_NACHRICHT_TEXT);
	}

	void addValueFromMapIfExists(Map<String, Object> sourceMap, String key, Map<String, Object> targetMap) {
		if (sourceMap.containsKey(key)) {
			targetMap.put(key, sourceMap.get(key));
		}
	}

	public void deleteBescheidDraft(Command command) {
		createCommandAfterValidation(command, bescheid -> buildDeleteItemCommand(command, bescheid));
	}

	private void createCommandAfterValidation(Command command, Function<AttachedItem, OzgCloudCommand> buildCommand) {
		var bescheid = remoteService.getItem(command.getRelationId());
		validateBescheidStatus(bescheid);
		commandService.createAndWaitUntilDone(buildCommand.apply(bescheid));
	}

	void validateBescheidStatus(AttachedItem bescheid) {
		var bescheidStatus = MapUtils.getString(bescheid.getItem(), Bescheid.FIELD_STATUS);
		if (Bescheid.Status.DRAFT.not(bescheidStatus)) {
			throw new TechnicalException("Bescheid draft with ID '%s' has an unexpected status: '%s'".formatted(bescheid.getId(), bescheidStatus));
		}
	}

	OzgCloudCommand buildDeleteItemCommand(Command command, AttachedItem bescheid) {
		return OzgCloudCommand.builder()
				.vorgangId(commandMapper.toOzgCloudVorgangId(command.getVorgangId()))
				.relationId(commandMapper.mapRelationId(bescheid.getId()))
				.relationVersion(bescheid.getVersion())
				.order(DELETE_ATTACHED_ITEM)
				.build();
	}

	Map<String, Object> buildItemMap(Status status) {
		var itemMap = new HashMap<String, Object>();
		itemMap.put(Bescheid.FIELD_STATUS, status.name());
		return itemMap;
	}

	public Collection<AttachedItem> findAllBescheid(VorgangId vorgangId) {
		return remoteService.findAllBescheid(vorgangId);
	}
}